import types
from collections import defaultdict, namedtuple
from uuid import uuid4
import time

from mock import Mock, patch
from tornado.httpclient import AsyncHTTPClient

from kafka_rest.client import KafkaRESTClient

class Callback(namedtuple('Callback', ['fn', 'args', 'kwargs'])):
    def run(self):
        return self.fn(*self.args, **self.kwargs)

class Later(namedtuple('Later', ['handle', 'seconds', 'callback'])):
    pass

def _mock_emitter(obj, event, *args, **kwargs):
    obj.event_mocks[event](*args, **kwargs)

class MockClient(KafkaRESTClient):
    @classmethod
    def with_basic_setup(cls):
        instance = cls(None, None, shutdown_timeout_seconds=0)
        instance.event_mocks = defaultdict(lambda: Mock())
        instance.registrar.emit = types.MethodType(_mock_emitter, instance)
        return instance

    def install_mock_io_loop(self):
        self.io_loop.stop()
        self.io_loop = TestingIOLoop()

        self.io_loop_patch = patch('kafka_rest.producer.IOLoop.current')
        self.io_loop_mock = self.io_loop_patch.start()
        self.io_loop_mock.return_value = self.io_loop

        self.producer.http_client = AsyncHTTPClient(io_loop=self.io_loop)

    def teardown_mock_io_loop(self):
        self.io_loop_patch.stop()

    def reset_mocks(self):
        for mock in self.event_mocks.values():
            mock.reset_mock()

    def mock_for(self, event):
        return self.event_mocks[event]

class TestingIOLoop(object):
    def __init__(self):
        self.callbacks = []
        self.laters = []

    def time(self):
        return time.time()

    def add_callback(self, cb, *args, **kwargs):
        self.callbacks.append(Callback(cb, args, kwargs))

    def call_later(self, seconds, cb, *args, **kwargs):
        handle = uuid4()
        self.laters.append(Later(handle, seconds, Callback(cb, args, kwargs)))
        return handle

    def add_timeout(self, deadline, cb, *args, **kwargs):
        return self.call_later(self.time() - deadline, cb, *args, **kwargs)

    def remove_timeout(self, handle):
        self.laters = filter(lambda later: later.handle != handle, self.laters)

    @property
    def next_callback(self):
        if self.callbacks:
            return self.callbacks[0]

    @property
    def next_later(self):
        if self.laters:
            return self.laters[0]

    def run_next_callback(self):
        if self.callbacks:
            self.callbacks[0].run()
            del self.callbacks[0]

    def pop_callback(self):
        if self.callbacks:
            return self.callbacks.pop(0)

    def pop_later(self):
        if self.laters:
            return self.laters.pop(0)

    @property
    def finished(self):
        return not bool(self.callbacks or self.laters)



from unittest import TestCase
from mock import Mock, ANY

from kafka_rest.events import EventRegistrar

def _broken_handler(*args, **kwargs):
    raise ValueError('wonk wonk')

class TestEventRegistrar(TestCase):
    def setUp(self):
        self.registrar = EventRegistrar()

    def test_emit_with_no_handlers(self):
        self.registrar.emit('test')

    def test_handlers_called_in_order(self):
        tracker = []
        handler_1 = lambda: tracker.append(1)
        handler_2 = lambda: tracker.append(2)
        self.registrar.register('test', handler_1)
        self.registrar.register('test', handler_2)

        self.registrar.emit('test')

        self.assertEqual(tracker, [1, 2])

    def test_handler_exception_emits_exception_event(self):
        self.registrar.register('test', _broken_handler)
        self.exception_mock = Mock()
        self.registrar.register('event_handler_exception', self.exception_mock)

        self.registrar.emit('test')

        self.exception_mock.assert_called_once_with('test', ANY)

    def test_broken_exception_event_does_not_infinite_loop(self):
        self.registrar.register('test', _broken_handler)
        self.registrar.register('event_handler_exception', _broken_handler)

        self.registrar.emit('test')



import sys
import logging
from threading import Thread
try:
    from queue import Queue, PriorityQueue, Full
except ImportError:
    from Queue import Queue, PriorityQueue, Full
from collections import defaultdict, namedtuple

from tornado.ioloop import IOLoop

from .events import EventRegistrar, DropReason
from .circuit_breaker import CircuitBreaker
from .producer import AsyncProducer
from .message import Message
from .exceptions import KafkaRESTShutdownException

logger = logging.getLogger('kafka_rest.client')

class KafkaRESTClient(object):
    def __init__(self, host, port, http_max_clients=10, max_queue_size_per_topic=5000,
                 flush_length_threshold=20, flush_time_threshold_seconds=20,
                 flush_max_batch_size=50, connect_timeout_seconds=10,
                 request_timeout_seconds=60, retry_base_seconds=2,
                 retry_max_attempts=10, retry_period_seconds=15,
                 response_5xx_circuit_breaker_trip_threshold=10,
                 response_5xx_circuit_breaker_trip_duration_seconds=300,
                 shutdown_timeout_seconds=2):
        self.host = host
        self.port = port
        self.http_max_clients = http_max_clients
        self.max_queue_size_per_topic = max_queue_size_per_topic
        self.flush_length_threshold = flush_length_threshold
        self.flush_time_threshold_seconds = flush_time_threshold_seconds
        self.flush_max_batch_size = flush_max_batch_size
        self.connect_timeout_seconds = connect_timeout_seconds
        self.request_timeout_seconds = request_timeout_seconds
        self.retry_base_seconds = retry_base_seconds
        # Includes the original send as an attempt, so set to 1 to disable retry
        self.retry_max_attempts = retry_max_attempts
        self.retry_period_seconds = retry_period_seconds
        # Circuit breaker prevents thrashing if we receive multiple transport
        # errors in a row, since this usually means we are experiencing some
        # sort of network problem and other requests are very likely to fail.
        # When this triggers, we wait a short duration before clearing the
        # breaker and attempting any further network operations.
        if response_5xx_circuit_breaker_trip_threshold is None:
            response_5xx_circuit_breaker_trip_threshold = sys.maxsize
        self.response_5xx_circuit_breaker_trip_threshold = response_5xx_circuit_breaker_trip_threshold
        self.response_5xx_circuit_breaker_trip_duration_seconds = response_5xx_circuit_breaker_trip_duration_seconds
        # On shutdown, last-ditch flush attempts are given this
        # request timeout after which they are considered failed
        self.shutdown_timeout_seconds = shutdown_timeout_seconds

        self.in_shutdown = False

        self.registrar = EventRegistrar()
        self.response_5xx_circuit_breaker = CircuitBreaker(self.response_5xx_circuit_breaker_trip_threshold,
                                                           self.response_5xx_circuit_breaker_trip_duration_seconds)
        self.message_queues = defaultdict(lambda: Queue(maxsize=max_queue_size_per_topic))
        self.retry_queues = defaultdict(lambda: PriorityQueue(maxsize=max_queue_size_per_topic))
        self.schema_cache = defaultdict(dict)
        self.io_loop = IOLoop()

        self.producer = AsyncProducer(self)
        self.io_loop.add_callback(self.producer._schedule_retry_periodically)

        self.producer_thread = Thread(target=self.io_loop.start)
        self.producer_thread.daemon = True
        self.producer_thread.start()
        logger.debug('Started producer background thread')

        logger.debug('Kafka REST async client initialized for {0}:{1}'.format(self.host, self.port))

    def produce(self, topic, value, value_schema, key=None, key_schema=None, partition=None):
        """Place this message on the appropriate topic queue for asynchronous
        emission."""
        if self.in_shutdown:
            raise KafkaRESTShutdownException('Client is in shutdown state, new events cannot be produced')

        if self.schema_cache[topic].get('value') is None:
            logger.debug('Storing initial value schema for topic {0} in schema cache: {1}'.format(topic, value_schema))
            self.schema_cache[topic]['value'] = value_schema
        if key_schema and self.schema_cache[topic].get('key') is None:
            logger.debug('Storing initial key schema for topic {0} in schema cache: {1}'.format(topic, key_schema))
            self.schema_cache[topic]['key'] = key_schema

        queue = self.message_queues[topic]
        message = Message(topic, value, key, partition, 0, 1)
        try:
            queue.put_nowait(message)
        except Full:
            logger.critical('Primary event queue is full for topic {0}, message {1} will be dropped'.format(topic, message))
            self.registrar.emit('drop_message', topic, message, DropReason.PRIMARY_QUEUE_FULL)
        else:
            self.registrar.emit('produce', message)
            self.io_loop.add_callback(self.producer.evaluate_queue, topic, queue)

    def shutdown(self, block=False):
        """Prohibit further produce requests and attempt to flush all events currently in
        the main and retry queues. After this attempt, all remaining events are made
        available to an event handler but will otherwise be dropped. The producer
        thread and IOLoop are also shut down. If block=True, this blocks until
        the producer thread is dead and the shutdown event has been handled."""
        logger.info('Client shutting down')
        self.in_shutdown = True
        self.io_loop.add_callback(self.producer.start_shutdown)

        if block:
            self.producer_thread.join()
            logger.info('Client completed shutdown')
        else:
            logger.info('Client shutting down asynchronously, will not block')



import time

class CircuitBreaker(object):
    def __init__(self, threshold, duration_seconds):
        self.threshold = threshold
        self.duration_seconds = duration_seconds
        self.failure_count = 0
        self._last_tripped = 0 # timestamp

    @property
    def tripped(self):
        return time.time() < self._last_tripped + self.duration_seconds

    def record_failure(self, value=1):
        self.failure_count += value
        if not self.tripped and self.failure_count >= self.threshold:
            self._last_tripped = time.time()

    def reset(self):
        self.failure_count = 0
        self._last_tripped = 0



from uuid import uuid4

from tornado.httpclient import HTTPRequest
from tornado.escape import json_encode
import avro.schema
from avro_json_serializer import AvroJsonSerializer

ERROR_CODES = {
    40401: 'Topic not found',
    40402: 'Partition not found',
    50001: 'Zookeeper error',
    50002: 'Kafka error',
    50003: 'Retriable Kafka error',
    50101: 'Only SSL endpoints found, but SSL not supported for invoked API'
}

# Confluent didn't seem very confident that 50002 would
# always be non-retriable when I asked them about it,
# so we'll keep it as retriable for now to be safe.
# See https://github.com/confluentinc/kafka-rest/issues/147
RETRIABLE_ERROR_CODES = set([50001, 50002, 50003])

def _encode_payload(schema_cache, topic, batch):
    value_schema = avro.schema.make_avsc_object(schema_cache[topic]['value'], avro.schema.Names())
    value_serializer = AvroJsonSerializer(value_schema)
    if schema_cache[topic].get('key') is not None:
        key_schema = avro.schema.make_avsc_object(schema_cache[topic]['key'], avro.schema.Names())
        key_serializer = AvroJsonSerializer(key_schema)

    body = {'records': [{'value': value_serializer.to_ordered_dict(message.value),
                         'key': key_serializer.to_ordered_dict(message.key) if message.key is not None else None,
                         'partition': message.partition}
                        for message in batch]}

    # The REST proxy's API requires us to double-encode the schemas.
    # Don't ask why, because I have no idea.
    if schema_cache[topic].get('value-id') is None:
        body['value_schema'] = json_encode(schema_cache[topic]['value'])
    else:
        body['value_schema_id'] = schema_cache[topic]['value-id']
    if schema_cache[topic].get('key') is not None:
        if schema_cache[topic].get('key-id') is None:
            body['key_schema'] = json_encode(schema_cache[topic]['key'])
        else:
            body['key_schema_id'] = schema_cache[topic]['key-id']

    return json_encode(body)

def request_for_batch(host, port, connect_timeout, request_timeout,
                      schema_cache, topic, batch):
    """Returns a Tornado HTTPRequest to the REST proxy
    representing the given message batch. This does not
    send the request, it just creates the object."""
    request = HTTPRequest('{0}:{1}/topics/{2}'.format(host, port, topic),
                          connect_timeout=connect_timeout,
                          request_timeout=request_timeout,
                          method='POST',
                          headers={'Accept': 'application/vnd.kafka.v1+json',
                                   'Content-Type': 'application/vnd.kafka.avro.v1+json'},
                          body=_encode_payload(schema_cache, topic, batch))

    # We also stick the message batch on the HTTPRequest object itself
    # so it is available to us when we handle the response. This is necessary
    # because individual messages can fail even if the overall request is
    # successful.
    request._topic = topic
    request._batch = batch

    # We also stick a unique ID on our request so we can keep track of
    # in-flight requests and the events contained in them
    request._id = uuid4()
    return request



from collections import namedtuple
import time

class Message(namedtuple('Message', ['topic', 'value', 'key', 'partition', 'retry_after_time', 'attempt_number'])):
    # Comparisons are defined for the sake of the retry PriorityQueues
    # and will not make sense for other comparisons
    def __eq__(self, other):
        return self.retry_after_time == other.retry_after_time

    def __ne__(self, other):
        return self.retry_after_time == other.retry_after_time

    def __lt__(self, other):
        return self.retry_after_time < other.retry_after_time

    def __le__(self, other):
        return self.retry_after_time <= other.retry_after_time

    def __gt__(self, other):
        return self.retry_after_time > other.retry_after_time

    def __ge__(self, other):
        return self.retry_after_time >= other.retry_after_time

    # Defined for the sake of tests
    def true_equals(self, other):
        if not isinstance(other, Message):
            return False
        for idx in range(len(self)):
            if self[idx] != other[idx]:
                return False
        return True

    def can_retry(self, client):
        return self.attempt_number < client.retry_max_attempts

    def for_retry(self, client):
        """Return a new Message object with the same data as this one
        but with incremented retry metadata."""
        backoff = client.retry_base_seconds ** self.attempt_number
        return Message(self.topic, self.value, self.key, self.partition,
                       time.time() + backoff, self.attempt_number + 1)



try:
    from queue import Queue
except ImportError:
    from Queue import Queue
from unittest import TestCase
import time

from mock import Mock

from .fixtures import MockClient
from kafka_rest.exceptions import KafkaRESTShutdownException
from kafka_rest.message import Message

class TestClient(TestCase):
    def setUp(self):
        self.client = MockClient.with_basic_setup()
        self.callback_mock = Mock(wraps=self.client.io_loop.add_callback)
        self.client.io_loop.add_callback = self.callback_mock

        self.test_schema = {
            'type': 'record',
            'name': 'test_driver',
            'fields': [
                {'name': 'val', 'type': 'int'}
            ]
        }
        self.test_value = {'val': 1}

    def tearDown(self):
        if not self.client.in_shutdown:
            self.client.shutdown(block=False)

    def test_initial_produce(self):
        self.client.produce('test_driver', self.test_value, self.test_schema)
        expected_message = Message('test_driver', self.test_value, None, None, 0, 1)

        self.assertEqual(self.client.schema_cache['test_driver']['value'], self.test_schema)
        self.assertEqual(None, self.client.schema_cache['test_driver'].get('key'))
        self.assertEqual(1, self.client.message_queues['test_driver'].qsize())
        self.client.mock_for('produce').assert_called_once_with(expected_message)
        self.callback_mock.assert_called_once_with(self.client.producer.evaluate_queue,
                                                   'test_driver',
                                                   self.client.message_queues['test_driver'])

    def test_produce_raises_in_shutdown(self):
        self.client.shutdown(block=True)
        self.assertRaises(KafkaRESTShutdownException, self.client.produce,
                          'test_driver', self.test_value, self.test_schema)

    def test_produce_with_full_queue(self):
        self.client.message_queues['test_driver'] = Queue(maxsize=1)
        self.client.message_queues['test_driver'].put('whatever')

        self.client.produce('test_driver', self.test_value, self.test_schema)
        expected_message = Message('test_driver', self.test_value, None, None, 0, 1)

        self.assertEqual(1, self.client.message_queues['test_driver'].qsize())
        self.assertTrue(self.client.mock_for('drop_message').called_with('test_driver',
                                                                         expected_message,
                                                                         'primary_queue_full'))

    def test_shutdown(self):
        self.client.shutdown()
        self.callback_mock.assert_any_call(self.client.producer.start_shutdown)

    def test_shutdown_blocking(self):
        self.client.shutdown(block=True)
        self.callback_mock.assert_any_call(self.client.producer.start_shutdown)
        self.assertFalse(self.client.producer_thread.is_alive())



"""These tests attempt to test the control flow of the producer as it
processes events. These are more complex than unit tests but they are
necessary because the producer's methods contain individual processing
logic (i.e. the flush topic method flushes a topic) but also are responsible
for registering other callbacks to continue operating the producer correctly
(i.e. flush topic is also responsible for resetting the flush timer)."""

from unittest import TestCase
from mock import Mock
import json

from kafka_rest.message import Message
from .fixtures import MockClient, Callback

class TestFlow(TestCase):
    def setUp(self):
        self.client = MockClient.with_basic_setup()
        self.client.install_mock_io_loop()
        self.producer = self.client.producer
        self.io_loop = self.client.io_loop

        self.test_schema = {
            'type': 'record',
            'name': 'test_driver',
            'fields': [
                {'name': 'val', 'type': 'int'}
            ]
        }
        self.test_value = {'val': 1}

        self.client.schema_cache['test_driver']['value'] = self.test_schema
        self.client.schema_cache['test_driver']['value-id'] = 1

        self.producer.inflight_requests = {1: None}

    def tearDown(self):
        if not self.client.in_shutdown:
            self.client.shutdown(block=False)
        self.client.teardown_mock_io_loop()

    def make_response(self, response_attrs=None, request_attrs=None):
        response = Mock()
        request = Mock()
        response.code = 200
        response.error = None
        response.body = json.dumps({'offsets': [{}]})
        request._id = 1
        request._topic = 'test_driver'
        request._batch = [Message('test_driver', self.test_value, None, None, 0, 1)]
        for k, v in (response_attrs or {}).items():
            if k == 'body':
                v = json.dumps(v)
            setattr(response, k, v)
        for k, v in (request_attrs or {}).items():
            setattr(request, k, v)
        response.request = request
        return response

    def test_produce_through_evaluate_no_flush(self):
        self.client.produce('test_driver', self.test_value, self.test_schema)

        self.assertEqual(self.io_loop.next_callback,
                         Callback(self.producer.evaluate_queue,
                                  ('test_driver', self.client.message_queues['test_driver']),
                                  {}))
        self.io_loop.run_next_callback()

        self.assertEqual(self.io_loop.next_later.seconds, self.client.flush_time_threshold_seconds)
        self.assertEqual(self.io_loop.next_later.callback,
                         Callback(self.producer._flush_topic,
                                  ('test_driver', 'time'),
                                  {}))

        self.io_loop.pop_later()
        self.assertTrue(self.io_loop.finished)

    def test_produce_through_evaluate_into_flush(self):
        self.client.flush_length_threshold = 0
        self.client.produce('test_driver', self.test_value, self.test_schema)

        self.assertEqual(self.io_loop.next_callback,
                         Callback(self.producer.evaluate_queue,
                                  ('test_driver', self.client.message_queues['test_driver']),
                                  {}))
        self.io_loop.run_next_callback()

        self.assertEqual(self.io_loop.next_callback.fn, self.producer._send_batch_produce_request)
        expected_message = Message('test_driver', self.test_value, None, None, 0, 1)
        self.assertTrue(self.io_loop.next_callback.args[1][0].true_equals(expected_message))


        self.assertEqual(self.io_loop.next_later.seconds, self.client.flush_time_threshold_seconds)
        self.assertEqual(self.io_loop.next_later.callback,
                         Callback(self.producer._flush_topic,
                                  ('test_driver', 'time'),
                                  {}))

        self.io_loop.pop_callback()
        self.io_loop.pop_later()

        self.assertTrue(self.io_loop.finished)

    def test_flush_through_send(self):
        m1 = Message('test_driver', {'val': 1}, None, None, 0, 1)
        queue = self.client.message_queues['test_driver']
        queue.put(m1)

        self.producer._flush_topic('test_driver', 'test')

        self.assertEqual(self.io_loop.next_callback.fn, self.producer._send_batch_produce_request)
        self.assertTrue(self.io_loop.next_callback.args[1][0].true_equals(m1))

        self.io_loop.run_next_callback()

        # Response handler callback from Tornado's HTTP client, this function
        # is defined inline so we can't test against it very well
        self.assertEqual(self.io_loop.next_callback.fn.__name__, 'handle_response')
        self.io_loop.pop_callback()

        self.assertEqual(self.io_loop.next_later.callback,
                         Callback(self.producer._flush_topic,
                                  ('test_driver', 'time'),
                                  {}))
        self.io_loop.pop_later()

        self.assertTrue(self.io_loop.finished)

    def test_response_through_success_no_failures(self):
        response = self.make_response()

        self.producer._handle_produce_response('test_driver', response)

        self.assertTrue(self.io_loop.finished)
        self.assertEqual(0, self.client.retry_queues['test_driver'].qsize())

    def test_response_through_success_with_some_nonretriable_failures(self):
        response = self.make_response(
            response_attrs={
                'body': {'offsets': [{'error_code': 1, 'message': 'Nonretriable'}]}})

        self.producer._handle_produce_response('test_driver', response)

        self.assertTrue(self.io_loop.finished)
        self.assertEqual(0, self.client.retry_queues['test_driver'].qsize())

    def test_response_through_success_with_some_retriable_failures(self):
        response = self.make_response(
            response_attrs={
                'body': {'offsets': [{'error_code': 2, 'message': 'Retriable'}]}})

        self.producer._handle_produce_response('test_driver', response)

        self.assertTrue(self.io_loop.finished)
        self.assertEqual(1, self.client.retry_queues['test_driver'].qsize())

    def test_response_through_success_with_transport_error(self):
        response = self.make_response(
            response_attrs={'code': 599})

        self.producer._handle_produce_response('test_driver', response)

        self.assertTrue(self.io_loop.finished)
        self.assertEqual(1, self.client.retry_queues['test_driver'].qsize())

    def test_response_through_success_with_retriable_full_error(self):
        response = self.make_response(
            response_attrs={'code': 500, 'body': {'error_code': 50001, 'message': 'Retriable'}})

        self.producer._handle_produce_response('test_driver', response)

        self.assertTrue(self.io_loop.finished)
        self.assertEqual(1, self.client.retry_queues['test_driver'].qsize())

    def test_response_through_success_with_nonretriable_full_error(self):
        response = self.make_response(
            response_attrs={'code': 500, 'body': {'error_code': 50101, 'message': 'Nonretriable'}})

        self.producer._handle_produce_response('test_driver', response)

        self.assertTrue(self.io_loop.finished)
        self.assertEqual(0, self.client.retry_queues['test_driver'].qsize())



import sys
from setuptools import setup, find_packages

from kafka_rest import __version__

install_requires = [
    'tornado>=4.0.0,<5.0.0',
    'avro_json_serializer>=0.4.1,<0.5.0',
    'avro==1.7.7'
]

setup(
    name='kafka-rest',
    version=__version__,
    description="Async client for Confluent's Kafka REST Proxy",
    url='https://github.com/gamechanger/kafka-rest',
    author='GameChanger',
    author_email='travis@gc.io',
    packages=find_packages(),
    install_requires=install_requires,
    tests_require=[
        'mock==1.3.0',
        'nose==1.3.7'
    ],
    test_suite="nose.collector",
    zip_safe=False
)



import time

from unittest import TestCase

from kafka_rest.circuit_breaker import CircuitBreaker

class TestCircuitBreaker(TestCase):
    def setUp(self):
        self.breaker = CircuitBreaker(2, 1)

    def test_not_tripped_at_init(self):
        self.assertFalse(self.breaker.tripped)

    def test_not_tripped_before_threshold(self):
        self.breaker.record_failure()
        self.assertEqual(1, self.breaker.failure_count)
        self.assertFalse(self.breaker.tripped)

    def test_tripped_after_two_incs(self):
        self.breaker.record_failure()
        self.breaker.record_failure()
        self.assertEqual(2, self.breaker.failure_count)
        self.assertTrue(self.breaker.tripped)

    def test_tripped_after_one_multi_inc(self):
        self.breaker.record_failure(2)
        self.assertEqual(2, self.breaker.failure_count)
        self.assertTrue(self.breaker.tripped)

    def test_tripped_after_one_overkill_multi_inc(self):
        self.breaker.record_failure(100)
        self.assertEqual(100, self.breaker.failure_count)
        self.assertTrue(self.breaker.tripped)

    def test_resets_after_duration_elapses(self):
        self.breaker.record_failure(2)
        self.assertTrue(self.breaker.tripped)
        time.sleep(1)
        self.assertFalse(self.breaker.tripped)

    def test_reset_after_init(self):
        self.breaker.reset()
        self.assertFalse(self.breaker.tripped)

    def test_reset_untrips(self):
        self.breaker.record_failure(2)
        self.assertTrue(self.breaker.tripped)
        self.breaker.reset()
        self.assertEqual(0, self.breaker.failure_count)
        self.assertFalse(self.breaker.tripped)
